// 1. Module definition 
module asteroid;

//2. Imports 
import 
	arc.scenegraph.all,
	arc.physics.shapes.circle,
	arc.math.routines,
	arc.types,
	arc.texture,
	arc.sound,
	arc.window,
	arc.particle;

import
	tango.core.Signal;

import 
	spacearea,
	scoreboard,
	laserbolt;
	
class AsteroidExplosion : Node, IDrawable, IAdvancable
{
	this(Node parent_)
	{
		super(parent_);
		
		scope col = makeTriangleRandom!(uniform1D)(Color.Red, Color.Yellow, Color(1., 0.5, 0.1));
		
		scope debrisvel = makeBallRandom!(risingLinear1D)(Point(0,0), 100.f);
		Particle debrisParticle()
		{
			Particle p;
			p.size = Size(6, 6);
			col.set(p.color);
			debrisvel.set(p.velocity);
			return p;
		}
		arc.particle.spawn(debris, 10u, &debrisParticle);
		
		scope glowpos = makeBallRandom!(uniform1D)(Point(0,0), 8f);
		scope glowvel = makeBallRandom!(fallingLinear1D)(Point(0,0), 16f);
		Particle glowParticle()
		{
			Particle p;
			p.size = Size(12,12);
			glowpos.set(p.position);
			glowvel.set(p.velocity);
			p.age = randomRange(0., 1.);
			col.set(p.color);
			return p;
		}		
		arc.particle.spawn(glow, 100u, &glowParticle);
	}
	
	void draw()
	{
		Texture drawtex = Texture("astbin/particle16.png");
	
		arc.particle.draw(glow, drawtex);
		arc.particle.draw(debris, drawtex);
	}
	
	void advance(arcfl msDt)
	{
		arcfl sDt = msDt / 1000.;
		
		arc.particle.age(debris, sDt);
		arc.particle.killOldAge(debris, 1.0f);
		arc.particle.fade(debris, 1.0f);
		arc.particle.move(debris, sDt);
		
		arc.particle.age(glow, sDt);
		arc.particle.killOldAge(glow, 2.0f);
		arc.particle.fade(glow, 2.0f);
		arc.particle.move(glow, sDt);		
		
		if(arc.particle.aliveCount(debris) == 0 && arc.particle.aliveCount(glow) == 0)
		{
			setParent(null);	
			sigEnd();
		}
	}
	
	Signal!() sigEnd;
	
private:
	struct Particle
	{
		Point position = Point(0,0);
		Point velocity;
		Size size = Size(32,32);
		Color color = Color.White;
		
		bool alive = true;
		arcfl age = 0.;
	}
	
	Particle[] debris;
	Particle[] glow;
}
	
///3. A single asteroid definition
class Asteroid : Node
{
public:
	//5. Create a random asteroid 
	static Asteroid createRandomAsteroid()
	{	
		// new asteroid 
		Asteroid ast = new Asteroid;	
		
		// set x and y locs at random locations 
		ast.physics.translation.x = randomRange(30,320);
		if(randomRange(0,1))
			ast.physics.translation.x += 450;
		
		ast.physics.translation.y = randomRange(30,220);
		if(randomRange(0,1))
			ast.physics.translation.y += 350;
		
		// set random velocities 
		ast.physics.velocity.x = (randomRange(0,1) - 0.5) / 10.;
		ast.physics.velocity.y = (randomRange(0,1) - 0.5) / 10.;
		
		// set random rotation amounts
		ast.physics.rotation = (randomRange(0,1) - 0.5) / 20.;
		
		// attach asteroid graphic to physics body
		ast.image = images[randomRange(0,2)];
		ast.image.makeNode(ast.physics);
				
		return ast;
	}
	
	//6. Create child asteroids 
	static Asteroid[] createChildAsteroids(Asteroid parent)
	{
		Asteroid[] children;

		// put three asteroids in the children
		children ~= [new Asteroid, new Asteroid, new Asteroid];
		
		foreach(ast; children)
		{
			// make the children 75% smaller 
			ast.image = parent.image;
			ast.image.makeNode(ast.physics);
			ast.physics.scale = parent.physics.scale * 0.75;
	
			// have same position, rotation, and velocities as parent 
			ast.physics.translation = parent.physics.translation;
			ast.physics.rotation = parent.physics.rotation;
			ast.physics.velocity = parent.physics.velocity;
			ast.physics.angularVelocity = parent.physics.angularVelocity;
		}
		
		// give the children some offset from the parent location and velocity 
		float offset = randomRange(0,100) * PI / 100;
		for(uint i = 0; i < 3; ++i)
		{
			Point dir = Point(1,0).rotate(offset + i * PI*2 / 3);
			children[i].physics.translation += 15 * children[i].physics.scale.x * dir;
			children[i].physics.velocity += randomRange(2,12) / 100. * dir;
		}
		
		return children;
	}
	
private:
	///7. if hit by laserbolt, split or explode
	void onCollide(Body a, Body b, arcfl msDt)
	{
		// only react to laser bolts
		if(cast(LaserBolt)b.getParent() !is null)
		{
			// 100 points per asteroid hit
			scoreBoard.addScore(100);
			
			// complimentary sound effects 
			explodeSound.stop();
			explodeSound.play();
			
			// if the asteroid is still large enough to split, split it
			if(a.scale.x > 1.5)
			{
				createChildAsteroids(this);
				destroy();
			}
			// otherwise, switch off collision and explode!
			else
			{
				a.passThroughBodies = true;
				a.triggerOtherCollisionSignals = false;
				a.triggerOwnCollisionSignals = false;
				physics.disconnectChildren();
				
				physics.angularVelocity = 0;
				auto explosion = new AsteroidExplosion(physics);
				explosion.sigEnd.attach(&destroy);
			}
		}
	}
	
	//8. will create asteroid with given texture and physics body and will create size
	// based on the amount of time that has passed
	this()
	{
		++count;
	
		// create new physics body for it
		physics = new Circle(this, 13, 50);
		// connect physics collision function 
		physics.sigSustainedPostCollide.attach(&onCollide);
		physics.scale = 2;
		physics.friction = 0.1;

    super(spacearea.area);
	}
	
	//9. Destroy the asteroid 
	void destroy()
	{
		setParent(null);
		--count;
	}

	//10. Asteroid variables
	Body physics;
	Image image;

	static Image[3] images;
	static count = 0;
	static LinearAnimation explosionAnimation;
	static Sound explodeSound;

	//4. Load asteroid resources 
	static void loadResources()
	{
		// 3 possible asteroid frames to display 
		Asteroid.images[0] = new Image(Texture("astbin/asteroid1.png"));
		Asteroid.images[1] = new Image(Texture("astbin/asteroid2.png"));
		Asteroid.images[2] = new Image(Texture("astbin/asteroid3.png"));
		
		// explode animation 
		explosionAnimation = new LinearAnimation(new Image(Texture("astbin/ship/explosion.png"), Rect(32,32)), 7, 100);
		
		// explode sound
		explodeSound = new Sound(new SoundFile("astbin/explosion.wav"));
		registerAutoProcessSound(explodeSound);
	}
	mixin onWindowOpen!(loadResources);	
}
